/*-
 * Copyright (c) 2010 Hans Petter Selasky. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/stdint.h>
#include <sys/stddef.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/systm.h>
#include <sys/conf.h>
#include <sys/kernel.h>
#include <sys/bus.h>
#include <sys/linker_set.h>
#include <sys/module.h>
#include <sys/lock.h>
#include <sys/mutex.h>
#include <sys/condvar.h>
#include <sys/sysctl.h>
#include <sys/unistd.h>
#include <sys/malloc.h>
#include <sys/priv.h>
#include <sys/uio.h>
#include <sys/poll.h>
#include <sys/sx.h>
#include <sys/queue.h>
#include <sys/fcntl.h>
#include <sys/vnode.h>
#include <sys/selinfo.h>

#include <machine/bus.h>

#include <vm/vm.h>
#include <vm/pmap.h>

#define	HAVE_CUSE_IOCTL

#include "cuse4bsd.h"

MODULE_VERSION(cuse4bsd, 1);

#define	NBUSY	((uint8_t *)1)

struct cuse_command;
struct cuse_server;
struct cuse_client;

struct cuse_client_command {
	TAILQ_ENTRY(cuse_client_command) entry;
	struct cuse_command sub;
	struct cuse_data_chunk chunk;
	struct sx sx;
	struct cv cv;
	struct thread *entered;
	struct cuse_client *client;
	uint32_t offset;
	uint32_t length;
	uint32_t buffer[CUSE_BUFFER_MAX / 4];
	int	got_signal;
	int	error;
	int	command;
};

struct cuse_memory {
	struct cuse_server *owner;
	uint8_t *virtaddr;
	uint32_t page_count;
	uint32_t in_use_by_server;
};

struct cuse_server_dev {
	TAILQ_ENTRY(cuse_server_dev) entry;
	struct cuse_server *server;
	struct cdev *kern_dev;
	struct cuse_dev *user_dev;
};

struct cuse_server {
	TAILQ_ENTRY(cuse_server) entry;
	TAILQ_HEAD(, cuse_client_command) head;
	TAILQ_HEAD(, cuse_server_dev) hdev;
	TAILQ_HEAD(, cuse_client) hcli;
	struct cv cv;
	struct selinfo selinfo;
	int	is_closing;
};

struct cuse_client {
	TAILQ_ENTRY(cuse_client) entry;
	struct cuse_client_command cmds[CUSE_CMD_MAX];
	struct cuse_server *server;
	struct cuse_server_dev *server_dev;
	int	fflags;
	int	is_closing;
};

static TAILQ_HEAD(, cuse_server) cuse_head;
static struct mtx cuse_mtx;
static struct cdev *cuse_dev;
static struct cuse_server *cuse_alloc_unit[CUSE_DEVICES_MAX];
static struct cuse_memory cuse_mem[CUSE_ALLOC_UNIT_MAX];

static d_open_t cuse_client_open;
static d_close_t cuse_client_close;
static d_ioctl_t cuse_client_ioctl;
static d_read_t cuse_client_read;
static d_write_t cuse_client_write;
static d_poll_t cuse_client_poll;
static d_mmap_t cuse_client_mmap;

static struct cdevsw cuse_client_devsw = {
	.d_version = D_VERSION,
	.d_open = cuse_client_open,
	.d_close = cuse_client_close,
	.d_ioctl = cuse_client_ioctl,
	.d_name = "cuse_client",
	.d_flags = D_TRACKCLOSE,
	.d_read = cuse_client_read,
	.d_write = cuse_client_write,
	.d_poll = cuse_client_poll,
	.d_mmap = cuse_client_mmap,
};

static d_open_t cuse_server_open;
static d_close_t cuse_server_close;
static d_ioctl_t cuse_server_ioctl;
static d_read_t cuse_server_read;
static d_write_t cuse_server_write;
static d_poll_t cuse_server_poll;
static d_mmap_t cuse_server_mmap;

static struct cdevsw cuse_server_devsw = {
	.d_version = D_VERSION,
	.d_open = cuse_server_open,
	.d_close = cuse_server_close,
	.d_ioctl = cuse_server_ioctl,
	.d_name = "cuse_server",
	.d_flags = D_TRACKCLOSE,
	.d_read = cuse_server_read,
	.d_write = cuse_server_write,
	.d_poll = cuse_server_poll,
	.d_mmap = cuse_server_mmap,
};

static void
cuse_lock(void)
{
	mtx_lock(&cuse_mtx);
}

static void
cuse_unlock(void)
{
	mtx_unlock(&cuse_mtx);
}

static void
cuse_cmd_lock(struct cuse_client_command *pccmd)
{
	sx_xlock(&pccmd->sx);
}

static void
cuse_cmd_unlock(struct cuse_client_command *pccmd)
{
	sx_xunlock(&pccmd->sx);
}

static void
cuse_kern_init(void *arg)
{
	TAILQ_INIT(&cuse_head);

	mtx_init(&cuse_mtx, "cuse-mtx", NULL, MTX_DEF);

	cuse_dev = make_dev(&cuse_server_devsw, 0,
	    UID_ROOT, GID_OPERATOR, 0600, "cuse");

	printf("Cuse4BSD v%d.%d.%d @ /dev/cuse\n",
	    (CUSE_VERSION >> 16) & 0xFF, (CUSE_VERSION >> 8) & 0xFF,
	    (CUSE_VERSION >> 0) & 0xFF);
}

SYSINIT(cuse_kern_init, SI_SUB_DEVFS, SI_ORDER_ANY, cuse_kern_init, 0);

static void
cuse_kern_uninit(void *arg)
{
	void *ptr;

	while (1) {

		cuse_lock();
		ptr = TAILQ_FIRST(&cuse_head);
		cuse_unlock();

		if (ptr == NULL)
			break;

		printf("Cuse4BSD: Please exit all /dev/cuse instances ...\n");

		pause("DRAIN", 2 * hz);
	}

	if (cuse_dev != NULL)
		destroy_dev(cuse_dev);

	mtx_destroy(&cuse_mtx);
}

SYSUNINIT(cuse_kern_uninit, SI_SUB_DEVFS, SI_ORDER_ANY, cuse_kern_uninit, 0);

static struct cuse_client_command *
cuse_server_find_command(struct cuse_server *pcs, struct thread *td)
{
	struct cuse_client *pcc;
	int n;

	TAILQ_FOREACH(pcc, &pcs->hcli, entry) {
		for (n = 0; n != CUSE_CMD_MAX; n++) {
			if (pcc->cmds[n].entered == td)
				return (&pcc->cmds[n]);
		}
	}
	return (NULL);
}

static void
cuse_str_filter(char *ptr)
{
	int c;

	while (((c = *ptr) != 0)) {

		if ((c >= 'a') && (c <= 'z')) {
			ptr++;
			continue;
		}
		if ((c >= 'A') && (c <= 'Z')) {
			ptr++;
			continue;
		}
		if ((c >= '0') && (c <= '9')) {
			ptr++;
			continue;
		}
		if ((c == '.') || (c == '_') || (c == '/')) {
			ptr++;
			continue;
		}
		*ptr = '_';

		ptr++;
	}
}

static int
cuse_convert_error(int error)
{
	;				/* indent fix */
	switch (error) {
	case CUSE_ERR_NONE:
		return (0);
	case CUSE_ERR_BUSY:
		return (EBUSY);
	case CUSE_ERR_WOULDBLOCK:
		return (EWOULDBLOCK);
	case CUSE_ERR_INVALID:
		return (EINVAL);
	case CUSE_ERR_NO_MEMORY:
		return (ENOMEM);
	case CUSE_ERR_FAULT:
		return (EFAULT);
	case CUSE_ERR_SIGNAL:
		return (EINTR);
	default:
		return (ENXIO);
	}
}

static int
cuse_server_wait_ready(struct cuse_client_command *pccmd)
{
	struct cuse_server *pcs = pccmd->client->server;

	int error;

	/* wait for ready state */
	while (pccmd->command != CUSE_CMD_NONE) {
		error = cv_wait_sig(&pccmd->cv, &cuse_mtx);
		if (cuse_server_find_command(pcs, curthread) != pccmd) {
			/* not opened */
			return (ENXIO);
		}
		if (error) {
			/* received signal */
			return (error);
		}
	}
	return (0);
}

static int
cuse_server_copy(struct cuse_client_command *pccmd, void *src,
    void *dst, int len, int do_copy_in)
{
	struct cuse_server *pcs = pccmd->client->server;
	int error;

	cuse_unlock();

	if (do_copy_in)
		error = copyin(src, dst, len);
	else
		error = copyout(src, dst, len);

	cuse_lock();

	if (cuse_server_find_command(pcs, curthread) != pccmd) {
		/* not opened */
		return (ENXIO);
	}
	return (error);
}

static void
cuse_server_free_memory(struct cuse_server *pcs)
{
	struct cuse_memory *mem;
	void *ptr[CUSE_ALLOC_UNIT_MAX];
	int n;

	for (n = 0; n != CUSE_ALLOC_UNIT_MAX; n++) {
		mem = &cuse_mem[n];

		ptr[n] = NULL;

		if (mem->owner != pcs)
			continue;

		if (mem->virtaddr == NULL)
			continue;

		if (mem->virtaddr == NBUSY)
			continue;

		ptr[n] = mem->virtaddr;

		mem->virtaddr = NULL;
		mem->owner = NULL;
	}

	cuse_unlock();
	for (n = 0; n != CUSE_ALLOC_UNIT_MAX; n++) {
		if (ptr[n] != NULL)
			free(ptr[n], M_TEMP);
	}
	cuse_lock();
}

static int
cuse_server_alloc_memory(struct cuse_server *pcs,
    struct cuse_memory *mem, uint32_t page_count)
{
	void *ptr;
	int error;

	cuse_lock();

	if (mem->virtaddr == NBUSY) {
		cuse_unlock();
		return (EBUSY);
	}
	if (mem->virtaddr != NULL) {
		if (mem->owner != pcs) {
			cuse_unlock();
			return (EBUSY);
		}
		if (mem->in_use_by_server != 0) {
			cuse_unlock();
			return (EBUSY);
		}
		if (mem->page_count == page_count) {
			mem->in_use_by_server = 1;
			cuse_unlock();
			return (0);
		}
		cuse_unlock();
		return (EBUSY);
	}
	memset(mem, 0, sizeof(*mem));

	mem->virtaddr = NBUSY;

	cuse_unlock();

	ptr = malloc(page_count * PAGE_SIZE, M_TEMP, M_WAITOK | M_ZERO);
	if (ptr == NULL)
		error = ENOMEM;
	else
		error = 0;

	cuse_lock();

	if (error) {
		mem->virtaddr = NULL;
		cuse_unlock();
		return (error);
	}
	mem->virtaddr = ptr;
	mem->page_count = page_count;
	mem->in_use_by_server = 1;
	mem->owner = pcs;
	cuse_unlock();

	return (0);
}

static void
cuse_client_send_command(struct cuse_client_command *pccmd,
    unsigned long data_ptr, unsigned long arg, int fflags, int ioflag)
{
	unsigned long cuse_fflags = 0;
	struct cuse_server *pcs;

	if (fflags & FREAD)
		cuse_fflags |= CUSE_FFLAG_READ;

	if (fflags & FWRITE)
		cuse_fflags |= CUSE_FFLAG_WRITE;

	if (ioflag & IO_NDELAY)
		cuse_fflags |= CUSE_FFLAG_NONBLOCK;

	cuse_lock();

	pccmd->sub.fflags = cuse_fflags;
	pccmd->sub.data_pointer = data_ptr;
	pccmd->sub.argument = arg;

	if (pccmd->entry.tqe_prev == NULL) {
		pcs = pccmd->client->server;
		TAILQ_INSERT_TAIL(&pcs->head, pccmd, entry);
		cv_signal(&pcs->cv);
	}
	cuse_unlock();
}

static void
cuse_client_got_signal(struct cuse_client_command *pccmd)
{
	struct cuse_server *pcs;

	pccmd->got_signal = 1;

	pccmd = &pccmd->client->cmds[CUSE_CMD_SIGNAL];

	if (pccmd->entry.tqe_prev == NULL) {
		pcs = pccmd->client->server;
		TAILQ_INSERT_TAIL(&pcs->head, pccmd, entry);
		cv_signal(&pcs->cv);
	}
}

static int
cuse_client_receive_command(struct cuse_client_command *pccmd,
    uint8_t *arg_ptr, uint32_t arg_len)
{
	struct cuse_server *pcs;
	int complete = 0;
	int error = 0;
	int server_cmd;
	uint32_t offset;
	uint32_t length;

	cuse_lock();

	pcs = pccmd->client->server;

	if (pcs->is_closing || pccmd->client->is_closing) {
		cuse_client_got_signal(pccmd);
		cuse_unlock();
		return (CUSE_ERR_OTHER);
	}
	while (!complete) {

		error = 0;

		while (pccmd->command == CUSE_CMD_NONE) {
			if (error != 0) {
				cv_wait(&pccmd->cv, &cuse_mtx);
			} else {
				error = cv_wait_sig(&pccmd->cv, &cuse_mtx);
				if (error != 0)
					cuse_client_got_signal(pccmd);
			}
			if (pcs->is_closing || pccmd->client->is_closing) {
				cuse_client_got_signal(pccmd);
				cuse_unlock();
				return (CUSE_ERR_OTHER);
			}
		}

		error = 0;

		server_cmd = pccmd->command;
		cuse_unlock();

		switch (server_cmd) {
		case CUSE_CMD_READ:
			if (pccmd->chunk.peer_ptr < CUSE_BUF_MIN_PTR) {
				error = CUSE_ERR_INVALID;
			} else if (pccmd->chunk.peer_ptr < CUSE_BUF_MAX_PTR) {
				offset = pccmd->chunk.peer_ptr - CUSE_BUF_MIN_PTR;
				length = pccmd->length;
				if (offset >= arg_len) {
					error = CUSE_ERR_INVALID;
					break;
				}
				if ((uint32_t)(offset + length) > arg_len) {
					error = CUSE_ERR_INVALID;
					break;
				}
				if ((uint32_t)(offset + length) < offset) {
					error = CUSE_ERR_INVALID;
					break;
				}
				if (arg_ptr == NULL) {
					error = CUSE_ERR_INVALID;
					break;
				}
				memcpy(pccmd->buffer, arg_ptr + offset, length);
				error = 0;
			} else {
				error = copyin(
				    (void *)pccmd->chunk.peer_ptr,
				    pccmd->buffer,
				    pccmd->length);
				if (error)
					error = CUSE_ERR_FAULT;
			}
			break;

		case CUSE_CMD_WRITE:
			if (pccmd->chunk.peer_ptr < CUSE_BUF_MIN_PTR) {
				error = CUSE_ERR_INVALID;
			} else if (pccmd->chunk.peer_ptr < CUSE_BUF_MAX_PTR) {
				offset = pccmd->chunk.peer_ptr - CUSE_BUF_MIN_PTR;
				length = pccmd->length;
				if (offset >= arg_len) {
					error = CUSE_ERR_INVALID;
					break;
				}
				if ((uint32_t)(offset + length) > arg_len) {
					error = CUSE_ERR_INVALID;
					break;
				}
				if ((uint32_t)(offset + length) < offset) {
					error = CUSE_ERR_INVALID;
					break;
				}
				if (arg_ptr == NULL) {
					error = CUSE_ERR_INVALID;
					break;
				}
				memcpy(arg_ptr + offset, pccmd->buffer, length);
				error = 0;
			} else {
				error = copyout(
				    pccmd->buffer,
				    (void *)pccmd->chunk.peer_ptr,
				    pccmd->length);
				if (error)
					error = CUSE_ERR_FAULT;
			}
			break;

		case CUSE_CMD_SYNC:
			error = pccmd->error;
			complete = 1;
			break;

		default:
			error = CUSE_ERR_INVALID;
			break;
		}

		cuse_lock();
		pccmd->error = error;
		pccmd->command = CUSE_CMD_NONE;
		cv_signal(&pccmd->cv);
	}

	cuse_unlock();
	return (error);
}

/*------------------------------------------------------------------------*
 *	CUSE4BSD SERVER PART
 *------------------------------------------------------------------------*/

static void
cuse_server_free_dev(struct cuse_server_dev *pcsd)
{
	struct cuse_server *pcs;
	struct cuse_client *pcc;
	int n;
	int error;

	pcs = pcsd->server;

	while (1) {

		error = 0;

		cuse_lock();
		TAILQ_FOREACH(pcc, &pcs->hcli, entry) {
			if (pcc->server_dev == pcsd) {
				pcc->is_closing = 1;
				for (n = 0; n != CUSE_CMD_MAX; n++) {
					cv_broadcast(&pcc->cmds[n].cv);
				}
				error = EBUSY;
			}
		}
		cuse_unlock();

		if (error)
			pause("CUSED", hz / 50);
		else
			break;
	}

	if (pcsd->kern_dev != NULL)
		destroy_dev(pcsd->kern_dev);

	free(pcsd, M_TEMP);
}

static void
cuse_server_free(void *arg)
{
	struct cuse_server *pcs = arg;
	struct cuse_server_dev *pcsd;
	struct cuse_client *pcc;
	int n;

	cuse_lock();

	pcs->is_closing = 1;

	while (TAILQ_FIRST(&pcs->hcli) != NULL) {

		TAILQ_FOREACH(pcc, &pcs->hcli, entry) {
			for (n = 0; n != CUSE_CMD_MAX; n++) {
				cv_signal(&pcc->cmds[n].cv);
			}
		}

		cuse_unlock();

		pause("CUSED", hz / 2);

		cuse_lock();
	}

	TAILQ_REMOVE(&cuse_head, pcs, entry);

	for (n = 0; n != CUSE_DEVICES_MAX; n++) {
		if (cuse_alloc_unit[n] == pcs)
			cuse_alloc_unit[n] = NULL;
	}

	while ((pcsd = TAILQ_FIRST(&pcs->hdev)) != NULL) {
		TAILQ_REMOVE(&pcs->hdev, pcsd, entry);
		cuse_unlock();
		cuse_server_free_dev(pcsd);
		cuse_lock();
	}

	cuse_server_free_memory(pcs);

	cuse_unlock();

	cv_destroy(&pcs->cv);

	free(pcs, M_TEMP);
}

static int
cuse_server_open(struct cdev *dev, int fflags, int devtype, struct thread *td)
{
	struct cuse_server *pcs;

	pcs = malloc(sizeof(*pcs), M_TEMP, M_WAITOK | M_ZERO);
	if (pcs == NULL)
		return (ENOMEM);

	TAILQ_INIT(&pcs->head);
	TAILQ_INIT(&pcs->hdev);
	TAILQ_INIT(&pcs->hcli);

	cv_init(&pcs->cv, "cuse-server-cv");

	cuse_lock();
	TAILQ_INSERT_TAIL(&cuse_head, pcs, entry);
	cuse_unlock();

	if (devfs_set_cdevpriv(pcs, &cuse_server_free))
		printf("cuse4bsd: Cannot set cdevpriv\n");

	return (0);
}

static int
cuse_server_close(struct cdev *dev, int fflag, int devtype, struct thread *td)
{
	struct cuse_server *pcs;
	int error;

	error = devfs_get_cdevpriv((void **)&pcs);
	if (error != 0)
		return (error);

	cuse_lock();
	pcs->is_closing = 1;
	cuse_unlock();

	return (0);
}

static int
cuse_server_read(struct cdev *dev, struct uio *uio, int ioflag)
{
	return (ENXIO);
}

static int
cuse_server_write(struct cdev *dev, struct uio *uio, int ioflag)
{
	return (ENXIO);
}

static int
cuse_server_ioctl(struct cdev *dev, unsigned long cmd,
    caddr_t data, int fflag, struct thread *td)
{
	struct cuse_server *pcs;
	int error;

	error = devfs_get_cdevpriv((void **)&pcs);
	if (error != 0)
		return (error);

	switch (cmd) {
		struct cuse_client_command *pccmd;
		struct cuse_client *pcc;
		struct cuse_command *pcmd;
		struct cuse_alloc_info *pai;
		struct cuse_create_dev *pcd;
		struct cuse_server_dev *pcsd;
		uint32_t temp;
		int first;
		int n;

	case CUSE_IOCTL_GET_COMMAND:
		pcmd = (void *)data;

		cuse_lock();

		while ((pccmd = TAILQ_FIRST(&pcs->head)) == NULL) {
			error = cv_wait_sig(&pcs->cv, &cuse_mtx);
			if (error) {
				cuse_unlock();
				return (error);
			}
		}

		TAILQ_REMOVE(&pcs->head, pccmd, entry);
		pccmd->entry.tqe_prev = NULL;

		pccmd->entered = curthread;

		*pcmd = pccmd->sub;

		cuse_unlock();

		break;

	case CUSE_IOCTL_SYNC_COMMAND:

		cuse_lock();
		while ((pccmd = cuse_server_find_command(pcs, curthread)) != NULL) {

			/* wait for ready state */
			error = cuse_server_wait_ready(pccmd);

			/* check for error */
			if (error)
				break;

			/* send sync command */
			pccmd->entered = NULL;
			pccmd->error = *(int *)data;
			pccmd->command = CUSE_CMD_SYNC;

			/* signal peer, if any */
			cv_signal(&pccmd->cv);
		}
		cuse_unlock();

		break;

	case CUSE_IOCTL_ALLOC_UNIT:

		cuse_lock();
		for (n = 0; n != CUSE_DEVICES_MAX; n++) {
			if (cuse_alloc_unit[n] == NULL) {
				cuse_alloc_unit[n] = pcs;
				break;
			}
		}
		cuse_unlock();

		if (n == CUSE_DEVICES_MAX)
			error = ENOMEM;
		else
			*(int *)data = n;
		break;

	case CUSE_IOCTL_FREE_UNIT:

		n = *(int *)data;

		cuse_lock();
		if ((n >= 0) &&
		    (n < CUSE_DEVICES_MAX) &&
		    (cuse_alloc_unit[n] == pcs))
			cuse_alloc_unit[n] = NULL;
		else
			error = EINVAL;
		cuse_unlock();
		break;

	case CUSE_IOCTL_ALLOC_MEMORY:

		pai = (void *)data;

		if (pai->alloc_nr >= CUSE_ALLOC_UNIT_MAX) {
			error = ENOMEM;
			break;
		}
		if (pai->page_count > CUSE_ALLOC_PAGES_MAX) {
			error = ENOMEM;
			break;
		}
		error = cuse_server_alloc_memory(pcs,
		    &cuse_mem[pai->alloc_nr], pai->page_count);
		break;

	case CUSE_IOCTL_FREE_MEMORY:
		pai = (void *)data;

		if (pai->alloc_nr >= CUSE_ALLOC_UNIT_MAX) {
			error = ENOMEM;
			break;
		}
		/* we trust the character device driver in this case */

		cuse_lock();
		if (cuse_mem[pai->alloc_nr].owner == pcs)
			cuse_mem[pai->alloc_nr].in_use_by_server = 0;
		else
			error = EINVAL;
		cuse_unlock();
		break;

	case CUSE_IOCTL_GET_SIG:

		cuse_lock();
		pccmd = cuse_server_find_command(pcs, curthread);

		if (pccmd != NULL) {
			n = pccmd->got_signal;
			pccmd->got_signal = 0;
		} else {
			n = 0;
		}
		cuse_unlock();

		*(int *)data = n;

		break;

	case CUSE_IOCTL_SET_PFH:

		cuse_lock();
		pccmd = cuse_server_find_command(pcs, curthread);

		if (pccmd != NULL) {
			pcc = pccmd->client;
			for (n = 0; n != CUSE_CMD_MAX; n++) {
				pcc->cmds[n].sub.per_file_handle = *(unsigned long *)data;
			}
		} else {
			error = ENXIO;
		}
		cuse_unlock();
		break;

	case CUSE_IOCTL_CREATE_DEV:

		error = priv_check(curthread, PRIV_DRIVER);
		if (error)
			break;

		pcd = (void *)data;

		/* filter input */

		pcd->devname[sizeof(pcd->devname) - 1] = 0;

		if (pcd->devname[0] == 0) {
			error = EINVAL;
			break;
		}
		cuse_str_filter(pcd->devname);

		pcd->permissions &= 0777;

		/* try to allocate a character device */

		pcsd = malloc(sizeof(*pcsd), M_TEMP, M_WAITOK | M_ZERO);

		if (pcsd == NULL) {
			error = ENOMEM;
			break;
		}
		pcsd->server = pcs;

		pcsd->user_dev = pcd->dev;

		pcsd->kern_dev = make_dev(&cuse_client_devsw,
		    0, pcd->user_id, pcd->group_id, pcd->permissions,
		    "%s", pcd->devname);

		if (pcsd->kern_dev == NULL) {
			free(pcsd, M_TEMP);
			error = ENOMEM;
			break;
		}
		pcsd->kern_dev->si_drv1 = pcsd;

		cuse_lock();
		TAILQ_INSERT_TAIL(&pcs->hdev, pcsd, entry);
		cuse_unlock();

		break;

	case CUSE_IOCTL_DESTROY_DEV:

		error = priv_check(curthread, PRIV_DRIVER);
		if (error)
			break;

		cuse_lock();

		error = EINVAL;

		pcsd = TAILQ_FIRST(&pcs->hdev);
		while (pcsd != NULL) {
			if (pcsd->user_dev == *(struct cuse_dev **)data) {
				TAILQ_REMOVE(&pcs->hdev, pcsd, entry);
				cuse_unlock();
				cuse_server_free_dev(pcsd);
				cuse_lock();
				error = 0;
				pcsd = TAILQ_FIRST(&pcs->hdev);
			} else {
				pcsd = TAILQ_NEXT(pcsd, entry);
			}
		}

		cuse_unlock();
		break;

	case CUSE_IOCTL_WRITE_DATA:

		cuse_lock();
		pccmd = cuse_server_find_command(pcs, curthread);
		if (pccmd == NULL) {
			/* not opened */
			cuse_unlock();
			return (ENXIO);
		}
		first = 1;
		temp = 0;

		while (1) {

			/* wait for ready state */
			error = cuse_server_wait_ready(pccmd);

			/* check for error */
			if (error)
				break;

			/* check if chunk should be initialised */
			if (first) {
				pccmd->chunk = *(struct cuse_data_chunk *)data;
				pccmd->error = 0;
				first = 0;
			}
			/* check for error */
			error = pccmd->error;
			if (error)
				break;

			/* post-increment peer pointer */
			pccmd->chunk.peer_ptr += temp;

			/* check if complete */
			if (pccmd->chunk.length == 0)
				break;

			/* cumpute maximum transfer length */
			temp = pccmd->chunk.length;
			if (temp > CUSE_BUFFER_MAX)
				temp = CUSE_BUFFER_MAX;

			/* copy data */
			error = cuse_server_copy(pccmd,
			    (void *)pccmd->chunk.local_ptr,
			    pccmd->buffer, temp, 1);

			/* check for error */
			if (error)
				break;

			/* update chunk state */
			pccmd->chunk.local_ptr += temp;
			pccmd->chunk.length -= temp;
			pccmd->length = temp;
			pccmd->command = CUSE_CMD_WRITE;

			/* signal peer */
			cv_signal(&pccmd->cv);
		}
		cuse_unlock();
		break;

	case CUSE_IOCTL_READ_DATA:

		cuse_lock();
		pccmd = cuse_server_find_command(pcs, curthread);
		if (pccmd == NULL) {
			/* not opened */
			cuse_unlock();
			return (ENXIO);
		}
		first = 1;

		while (1) {

			/* wait for ready state */
			error = cuse_server_wait_ready(pccmd);

			/* check for error */
			if (error)
				break;

			/* check if chunk should be initialised */
			if (first) {
				pccmd->chunk = *(struct cuse_data_chunk *)data;
				pccmd->error = 0;
				first = 0;
			}
			error = pccmd->error;

			/* check for error */
			if (error)
				break;

			/* check if complete */
			if (pccmd->chunk.length == 0)
				break;

			/* find maximum length */
			temp = pccmd->chunk.length;
			if (temp > CUSE_BUFFER_MAX)
				temp = CUSE_BUFFER_MAX;

			pccmd->length = temp;
			pccmd->command = CUSE_CMD_READ;

			/* signal peer */
			cv_signal(&pccmd->cv);

			/* wait for ready state */
			error = cuse_server_wait_ready(pccmd);

			/* check for error */
			if (error)
				break;

			/* copy data */
			error = cuse_server_copy(pccmd, pccmd->buffer,
			    (void *)pccmd->chunk.local_ptr, temp, 0);

			/* check for error */
			if (error)
				break;

			/* update chunk state */
			pccmd->chunk.local_ptr += temp;
			pccmd->chunk.peer_ptr += temp;
			pccmd->chunk.length -= temp;
		}
		cuse_unlock();
		break;

	case CUSE_IOCTL_SELWAKEUP:
		selwakeup(&pcs->selinfo);
		break;

	default:
		error = ENXIO;
		break;
	}
	return (error);
}

static int
cuse_server_poll(struct cdev *dev, int events, struct thread *td)
{
	return (events & (POLLHUP | POLLPRI | POLLIN |
	    POLLRDNORM | POLLOUT | POLLWRNORM));
}

static int
#if (__FreeBSD_version >= 900000)
cuse_server_mmap(struct cdev *dev, vm_ooffset_t offset, vm_paddr_t *paddr, int nprot, vm_memattr_t *memattr)
#else
cuse_server_mmap(struct cdev *dev, vm_offset_t offset, vm_paddr_t *paddr, int nprot)
#endif
{
	uint32_t page_nr = offset / PAGE_SIZE;
	uint32_t alloc_nr = page_nr / CUSE_ALLOC_PAGES_MAX;
	struct cuse_server *pcs;
	struct cuse_memory *mem;
	uint8_t *ptr;
	int error;

	error = devfs_get_cdevpriv((void **)&pcs);
	if (error != 0)
		pcs = NULL;

	if (alloc_nr >= CUSE_ALLOC_UNIT_MAX) {
		return (ENOMEM);
	}
	cuse_lock();
	mem = &cuse_mem[alloc_nr];

	if ((pcs != NULL) && (mem->owner != pcs)) {
		cuse_unlock();
		return (EINVAL);
	}
	if (mem->virtaddr == NULL) {
		cuse_unlock();
		return (ENOMEM);
	}
	if (mem->virtaddr == NBUSY) {
		cuse_unlock();
		return (ENOMEM);
	}
	page_nr %= CUSE_ALLOC_PAGES_MAX;

	if (page_nr >= mem->page_count) {
		cuse_unlock();
		return (ENXIO);
	}
	ptr = mem->virtaddr + (page_nr * PAGE_SIZE);
	cuse_unlock();

	*paddr = vtophys(ptr);

	return (0);
}

/*------------------------------------------------------------------------*
 *	CUSE4BSD CLIENT PART
 *------------------------------------------------------------------------*/
static void
cuse_client_free(void *arg)
{
	struct cuse_client *pcc = arg;
	struct cuse_client_command *pccmd;
	int n;
	int error;

	cuse_lock();

	pcc->is_closing = 1;

	TAILQ_REMOVE(&pcc->server->hcli, pcc, entry);

	cuse_unlock();

	error = 0;

	cuse_lock();
	for (n = 0; n != CUSE_CMD_MAX; n++) {

		pccmd = &pcc->cmds[n];

		if (pccmd->entry.tqe_prev != NULL) {
			TAILQ_REMOVE(&pcc->server->head, pccmd, entry);
			pccmd->entry.tqe_prev = NULL;
		}
		if (pccmd->entered != NULL)
			error = EBUSY;

		/* wakeup all sleepers */
		cv_broadcast(&pccmd->cv);
	}
	cuse_unlock();

	/* make sure all server threads have returned */
	if (error != 0)
		pause("DRAIN", hz / 50);

	for (n = 0; n != CUSE_CMD_MAX; n++) {

		pccmd = &pcc->cmds[n];

		sx_destroy(&pccmd->sx);
		cv_destroy(&pccmd->cv);
	}

	free(pcc, M_TEMP);
}

static int
cuse_client_open(struct cdev *dev, int fflags, int devtype, struct thread *td)
{
	struct cuse_client_command *pccmd;
	struct cuse_server_dev *pcsd;
	struct cuse_client *pcc;
	int error;
	int n;

	pcsd = dev->si_drv1;
	if (pcsd == NULL)
		return (EINVAL);

	pcc = malloc(sizeof(*pcc), M_TEMP, M_WAITOK | M_ZERO);
	if (pcc == NULL)
		return (ENOMEM);

	pcc->fflags = fflags;
	pcc->server_dev = pcsd;
	pcc->server = pcsd->server;

	for (n = 0; n != CUSE_CMD_MAX; n++) {

		pccmd = &pcc->cmds[n];

		pccmd->sub.dev = pcsd->user_dev;
		pccmd->sub.command = n;
		pccmd->client = pcc;

		sx_init(&pccmd->sx, "cuse-client-sx");
		cv_init(&pccmd->cv, "cuse-client-cv");
	}

	if (devfs_set_cdevpriv(pcc, &cuse_client_free))
		printf("cuse4bsd: Cannot set cdevpriv\n");

	cuse_lock();

	/* cuse_client_free() assumes that the client is listed somewhere! */
	/* always enqueue */
	TAILQ_INSERT_TAIL(&pcc->server->hcli, pcc, entry);

	if (pcc->server->is_closing) {
		error = EINVAL;
	} else {
		error = 0;
	}
	cuse_unlock();

	if (error)
		return (error);

	pccmd = &pcc->cmds[CUSE_CMD_OPEN];

	cuse_cmd_lock(pccmd);

	cuse_client_send_command(pccmd, 0, 0, pcc->fflags, 0);

	error = cuse_client_receive_command(pccmd, 0, 0);
	if (error < 0) {
		error = cuse_convert_error(error);
	} else {
		error = 0;
	}

	cuse_cmd_unlock(pccmd);

	return (error);
}

static int
cuse_client_close(struct cdev *dev, int fflag, int devtype, struct thread *td)
{
	struct cuse_client_command *pccmd;
	struct cuse_client *pcc;
	int error;

	error = devfs_get_cdevpriv((void **)&pcc);
	if (error != 0)
		return (0);

	pccmd = &pcc->cmds[CUSE_CMD_CLOSE];

	cuse_cmd_lock(pccmd);

	cuse_client_send_command(pccmd, 0, 0, pcc->fflags, 0);

	error = cuse_client_receive_command(pccmd, 0, 0);

	cuse_cmd_unlock(pccmd);

	return (0);
}

static int
cuse_client_read(struct cdev *dev, struct uio *uio, int ioflag)
{
	struct cuse_client_command *pccmd;
	struct cuse_client *pcc;
	int error;
	int len;

	error = devfs_get_cdevpriv((void **)&pcc);
	if (error != 0)
		return (error);

	pccmd = &pcc->cmds[CUSE_CMD_READ];

	if (uio->uio_segflg != UIO_USERSPACE)
		return (EINVAL);

	uio->uio_segflg = UIO_NOCOPY;

	cuse_cmd_lock(pccmd);

	while (uio->uio_resid != 0) {

		len = uio->uio_iov->iov_len;

		cuse_client_send_command(pccmd,
		    (unsigned long)uio->uio_iov->iov_base,
		    (unsigned long)len, pcc->fflags, ioflag);

		error = cuse_client_receive_command(pccmd, 0, 0);
		if (error < 0) {
			error = cuse_convert_error(error);
			break;
		} else if (error == len) {
			error = uiomove(NULL, error, uio);
			if (error)
				break;
		} else {
			error = uiomove(NULL, error, uio);
			break;
		}
	}

	cuse_cmd_unlock(pccmd);

	return (error);
}

static int
cuse_client_write(struct cdev *dev, struct uio *uio, int ioflag)
{
	struct cuse_client_command *pccmd;
	struct cuse_client *pcc;
	int error;
	int len;

	error = devfs_get_cdevpriv((void **)&pcc);
	if (error != 0)
		return (error);

	pccmd = &pcc->cmds[CUSE_CMD_WRITE];

	if (uio->uio_segflg != UIO_USERSPACE)
		return (EINVAL);

	uio->uio_segflg = UIO_NOCOPY;

	cuse_cmd_lock(pccmd);

	while (uio->uio_resid != 0) {

		len = uio->uio_iov->iov_len;

		cuse_client_send_command(pccmd,
		    (unsigned long)uio->uio_iov->iov_base,
		    (unsigned long)len, pcc->fflags, ioflag);

		error = cuse_client_receive_command(pccmd, 0, 0);
		if (error < 0) {
			error = cuse_convert_error(error);
			break;
		} else if (error == len) {
			error = uiomove(NULL, error, uio);
			if (error)
				break;
		} else {
			error = uiomove(NULL, error, uio);
			break;
		}
	}

	cuse_cmd_unlock(pccmd);

	return (error);
}

int
cuse_client_ioctl(struct cdev *dev, unsigned long cmd,
    caddr_t data, int fflag, struct thread *td)
{
	struct cuse_client_command *pccmd;
	struct cuse_client *pcc;
	int error;
	int len;

	error = devfs_get_cdevpriv((void **)&pcc);
	if (error != 0)
		return (error);

	len = IOCPARM_LEN(cmd);
	if (len > CUSE_BUFFER_MAX)
		return (ENOMEM);

	pccmd = &pcc->cmds[CUSE_CMD_IOCTL];

	cuse_cmd_lock(pccmd);

	/* 
	 * When the ioctl-length is zero drivers can pass information
	 * through the data pointer of the ioctl. Make sure this
	 * information is forwarded to the driver.
	 */

	cuse_client_send_command(pccmd,
	    (len == 0) ? *(long *)data : CUSE_BUF_MIN_PTR,
	    (unsigned long)cmd, pcc->fflags,
	    (fflag & O_NONBLOCK) ? IO_NDELAY : 0);

	error = cuse_client_receive_command(pccmd, data, len);
	if (error < 0) {
		error = cuse_convert_error(error);
	} else {
		error = 0;
	}

	cuse_cmd_unlock(pccmd);

	return (error);
}

static int
cuse_client_poll(struct cdev *dev, int events, struct thread *td)
{
	struct cuse_client_command *pccmd;
	struct cuse_client *pcc;
	unsigned long temp;
	int error;
	int revents;

	error = devfs_get_cdevpriv((void **)&pcc);
	if (error != 0)
		return (POLLNVAL);

	temp = 0;

	if (events & (POLLPRI | POLLIN | POLLRDNORM))
		temp |= CUSE_POLL_READ;

	if (events & (POLLOUT | POLLWRNORM))
		temp |= CUSE_POLL_WRITE;

	if (events & POLLHUP)
		temp |= CUSE_POLL_ERROR;

	pccmd = &pcc->cmds[CUSE_CMD_POLL];

	cuse_cmd_lock(pccmd);

	cuse_client_send_command(pccmd,
	    0, temp, pcc->fflags, IO_NDELAY);

	error = cuse_client_receive_command(pccmd, 0, 0);
	if (error < 0) {
		revents = POLLNVAL;
	} else {
		revents = 0;
		if (error & CUSE_POLL_READ)
			revents |= (events & (POLLPRI | POLLIN | POLLRDNORM));
		if (error & CUSE_POLL_WRITE)
			revents |= (events & (POLLOUT | POLLWRNORM));
		if (error & CUSE_POLL_ERROR)
			revents |= (events & POLLHUP);
		if ((revents == 0) && (temp != 0))
			selrecord(curthread, &pcc->server->selinfo);
	}

	cuse_cmd_unlock(pccmd);

	return (revents);
}

static int
#if (__FreeBSD_version >= 900000)
cuse_client_mmap(struct cdev *dev, vm_ooffset_t offset, vm_paddr_t *paddr, int nprot, vm_memattr_t *memattr)
#else
cuse_client_mmap(struct cdev *dev, vm_offset_t offset, vm_paddr_t *paddr, int nprot)
#endif
{
	uint32_t page_nr = offset / PAGE_SIZE;
	uint32_t alloc_nr = page_nr / CUSE_ALLOC_PAGES_MAX;
	struct cuse_server *pcs;
	struct cuse_client *pcc;
	struct cuse_memory *mem;
	uint8_t *ptr;
	int error;

	error = devfs_get_cdevpriv((void **)&pcc);
	if (error != 0) {
		pcc = NULL;
		pcs = NULL;
	} else {
		pcs = pcc->server;
	}

	if (alloc_nr >= CUSE_ALLOC_UNIT_MAX) {
		return (ENOMEM);
	}
	cuse_lock();
	mem = &cuse_mem[alloc_nr];

	if ((pcs != NULL) && (mem->owner != pcs)) {
		cuse_unlock();
		return (EINVAL);
	}
	if (mem->virtaddr == NULL) {
		cuse_unlock();
		return (ENOMEM);
	}
	if (mem->virtaddr == NBUSY) {
		cuse_unlock();
		return (ENOMEM);
	}
	page_nr %= CUSE_ALLOC_PAGES_MAX;

	if (page_nr >= mem->page_count) {
		cuse_unlock();
		return (ENXIO);
	}
	ptr = mem->virtaddr + (page_nr * PAGE_SIZE);
	cuse_unlock();

	*paddr = vtophys(ptr);

	return (0);
}
